
/*
 * SA11x0 + UCB1x00 Touch Screen Driver Version 0.2
 * Put together by Tak-Shing Chan <tchan.rd@idthk.com>
 *
 * 90% derived from the Touch screen driver for Tifon
 * Copyright 1999 Peter Danielsson
 *
 * Codec routines derived from Itsy's Touchscreen driver
 * Copyright (c) Compaq Computer Corporation, 1998, 1999
 * Author: Larry Brakmo.
 *
 * Sample filtering derived from Diagman (ucb1x00Touch.c)
 * Copyright (C) 1999 Intel Corp.
 *
 * IOCTL's and jitter elimination derived from ADS Touchscreen driver
 * Copyright (c) 2000 Century Software, Inc.
 *
 * iPAQ emulation derived from Driver for the h3600 Touch Screen
 * Copyright 2000 Compaq Computer Corporation.
 * Author: Charles Flynn.
 *
 * Add Some function to let it more flexible
 * Allen_cheng@xlinux.com
 * chester@linux.org.tw
 */

/*
 * INSTRUCTIONS
 * To use this driver, simply ``mknod /dev/h3600_ts c 11 0'' and
 * use it as if it were an iPAQ.  TODO: iPAQ-compatible IOCTL's.
 * 
 *
 */
/* #define DPRINTK(x...) printk(KERN_NOTICE x) */
#define DPRINTK(x...)

#include <linux/types.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <asm/hardware.h>
#include <asm/irq.h>
#include <linux/string.h>
/* Audio driver includes */
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/interrupt.h>
#include <linux/errno.h>
#include <linux/sound.h>
#include <linux/soundcard.h>

#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/hardware.h>
#include <asm/semaphore.h>
#include <asm/dma.h>

#include "ucb1200_ts.h"

#define DELAY_DURATION 200

/* From Compaq's Touch Screen Specification version 0.2 (draft) */
typedef struct {
    short pressure;
    short x;
    short y;
    short millisecs;
} TS_EVENT;

static int raw_max_x=896, raw_max_y=886, res_x, res_y, raw_min_x=187,
raw_min_y=127, xyswap=0;

/*
 raw_max_x = 896;
 raw_max_y = 886;
 raw_min_x = 187;
 raw_min_y = 127;
 xyswap = 0;
*/

static int cal_ok, x_rev, y_rev;

static char *dev_id = "ucb1200";

static wait_queue_head_t queue;
static struct timer_list timer;
static int Reg10;
#define PRESSED 0
#define P_DONE 1
#define X_DONE 2
#define Y_DONE 3
#define RELEASED 4
#define BUFSIZE 128
#define XLIMIT 160
#define YLIMIT 160
static int state, head, tail, sample;
static TS_EVENT cur_data, samples[3], buf[BUFSIZE];
static struct fasync_struct *fasync;
static unsigned long in_timehandle = 0;

/* Allen Add */
static void ts_clear(void);
static void print_par(void);

/* Audio Codec Stuff */

#include <linux/sound.h>
#include <linux/soundcard.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/hardware.h>
#include <asm/semaphore.h>
#include <asm/dma.h>
#include "ucb1200.h"

#define AUDIO_NAME		"UCB1300"
#define AUDIO_NAME_VERBOSE	"UCB1300 audio driver"
#define AUDIO_VERSION_STRING	"version 0.1"

#define AUDIO_RATE_DEFAULT	44100
#define AUDIO_NBFRAGS_DEFAULT	8
#define AUDIO_FRAGSIZE_DEFAULT	8192



typedef struct {
	int size;		/* buffer size */
	char *start;		/* points to actual buffer */
	dma_addr_t dma_addr;	/* physical buffer address */
	struct semaphore sem;	/* down before touching the buffer */
	int master;		/* master owner for buffer allocation */
} audio_buf_t;

typedef struct {
	audio_buf_t *buffers;	/* pointer to audio buffer structures */
	audio_buf_t *buf;	/* current buffer used by read/write */
	u_int buf_idx;		/* index for the pointer above... */
	u_int fragsize;		/* fragment i.e. buffer size */
	u_int nbfrags;		/* nbr of fragments i.e. buffers */
	dmach_t dma_ch;		/* DMA channel ID */
} audio_stream_t;

static audio_stream_t ucb1300_output_stream;
static audio_stream_t ucb1300_input_stream;

static long int op_dma_chunk=0;

#define NEXT_BUF(_s_,_b_) { \
	(_s_)->_b_##_idx++; \
	(_s_)->_b_##_idx %= (_s_)->nbfrags; \
	(_s_)->_b_ = (_s_)->buffers + (_s_)->_b_##_idx; }


/* Current specs for incoming audio data */
static u_int audio_rate;
static u_int audio_fragsize;
static u_int audio_nbfrags;


static int audio_refcount;	/* nbr of concurrent open() */
static int audio_dev_dsp;	/* registered ID for DSP device */
static int audio_dev_mixer;	/* registered ID for mixer device */

static inline void audio_assabet_ucb1300_init(void);

/* Audio Codec Stuff */

/* Useful operations */
static inline ushort codec_read(ushort addr)
{
    ulong flags;
    ushort data;

    save_flags_cli(flags);
    while (!(Ser4MCSR & MCSR_CWC));
    Ser4MCDR2 = ((addr & 0xf) << FShft(MCDR2_ADD)) | MCDR2_Rd;
    while (!(Ser4MCSR & MCSR_CRC));
    data = Ser4MCDR2 & 0xffff;
    restore_flags(flags);
    return data;
}

static inline void codec_write(ushort addr, ushort data)
{
    ulong flags;

    save_flags_cli(flags);
    while (!(Ser4MCSR & MCSR_CWC));
    Ser4MCDR2 = ((addr & 0xf) << FShft(MCDR2_ADD)) | MCDR2_Wr | (data & 0xffff);
    restore_flags(flags);
}

static inline void ts_clear_codec_int(void)
{
   int s;

/* Clear interrupts at codec  */
   /* Vivek */
   /* if ((s = codec_read(CODEC_REG_INT_STATUS) & (TSPX_INT | ADC_INT))) */
   if ((s = codec_read(CODEC_REG_INT_STATUS) & (TSPX_INT | TSMX_INT | ADC_INT)))
   {
      codec_write(CODEC_REG_INT_STATUS, 0);
      codec_write(CODEC_REG_INT_STATUS, s);
   }
}

static inline void set_read_x_pos(void)
{
/* See Philips' AN809 for an explanation of the pressure mode switch */
   codec_write(CODEC_REG_TS_CTL, TSPX_POW | TSMX_GND | TSC_MODE_PRESSURE | TSC_BIAS_ENA);

   codec_write(CODEC_REG_TS_CTL, TSPX_POW | TSMX_GND | TSC_MODE_POSITION | TSC_BIAS_ENA);

   codec_write(CODEC_REG_ADC_CTL, Reg10 = ADC_INPUT_TSPY | ADC_ENA | ADC_SYNC_ENA);

   udelay(DELAY_DURATION);

   codec_write(CODEC_REG_ADC_CTL, Reg10 | ADC_START);

   codec_write(CODEC_REG_IO_DATA, codec_read(CODEC_REG_IO_DATA) | 0x200);

   codec_write(CODEC_REG_ADC_CTL, Reg10);

   codec_write(CODEC_REG_IO_DATA, codec_read(CODEC_REG_IO_DATA) & ~0x200);
}

static inline void set_read_y_pos(void)
{
   codec_write(CODEC_REG_TS_CTL, TSPY_POW | TSMY_GND | TSC_MODE_PRESSURE | TSC_BIAS_ENA);

   codec_write(CODEC_REG_TS_CTL, TSPY_POW | TSMY_GND | TSC_MODE_POSITION | TSC_BIAS_ENA);

   codec_write(CODEC_REG_ADC_CTL, Reg10 = ADC_INPUT_TSPX | ADC_ENA | ADC_SYNC_ENA);


   udelay(DELAY_DURATION);

   codec_write(CODEC_REG_ADC_CTL, Reg10 | ADC_START);

   codec_write(CODEC_REG_IO_DATA, codec_read(CODEC_REG_IO_DATA) | 0x200);

   codec_write(CODEC_REG_ADC_CTL, Reg10);

   codec_write(CODEC_REG_IO_DATA, codec_read(CODEC_REG_IO_DATA) & ~0x200);
}

static inline void set_read_pressure(void)
{
   codec_write(CODEC_REG_TS_CTL, TSPX_POW | TSMX_POW | TSPY_GND | TSMY_GND | TSC_MODE_PRESSURE | TSC_BIAS_ENA);

   codec_write(CODEC_REG_TS_CTL, TSPX_POW | TSMX_POW | TSPY_GND | TSMY_GND | TSC_MODE_PRESSURE | TSC_BIAS_ENA);

   codec_write(CODEC_REG_ADC_CTL, Reg10 = ADC_INPUT_TSPX | ADC_ENA | ADC_SYNC_ENA);

   udelay(DELAY_DURATION);

   codec_write(CODEC_REG_ADC_CTL, Reg10 | ADC_START);

   codec_write(CODEC_REG_IO_DATA, codec_read(CODEC_REG_IO_DATA) | 0x200);

   codec_write(CODEC_REG_ADC_CTL, Reg10);

   codec_write(CODEC_REG_IO_DATA, codec_read(CODEC_REG_IO_DATA) & ~0x200);
}

static int read_adc(void)
{
   int v;

   while (!((v = codec_read(CODEC_REG_ADC_DATA)) & ADC_DAT_VAL));
      return GET_ADC_DATA(v);
}

static int sa1100_ts_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
   int ret_data = 0;

/* Microwindows style (should change to TS_CAL when the specification is ready) */
   switch (cmd)
   {
      case 3:
	 raw_max_x = arg;
	 break;
      case 4:
	 raw_max_y = arg;
	 break;
      case 5:
	 res_x = arg;
	 break;
      case 6:
	 res_y = arg;
	 break;
      case 10:
	 raw_min_x = arg;
	 break;
      case 11:
	 raw_min_y = arg;
	 break;
      case 12:
/* New attribute for portrait modes */
	 xyswap = arg;
/* Allen Add */
      case 13:	       /* 0 = Enable calibration ; 1 = Calibration OK */
	 cal_ok = arg;
      case 14:	       /* Clear all buffer data */
	 ts_clear();
	 break;
      case 15:	       /* X axis reversed setting */
	 x_rev = arg;
	 break;
      case 16:	       /* X axis reversed setting */
	 y_rev = arg;
	 break;
      case 17:	       /* Clear all buffer data */
	 print_par();
	 break;
/* Allen */
   }

   return 0;
}

static void ts_clear(void)
{
   int i;

   for (i=0; i < BUFSIZE; i++)
   {
       buf[i].pressure=(short)NULL;
       buf[i].x=(int)NULL;
       buf[i].y=(int)NULL;
       buf[i].millisecs=(int)NULL;
   }

   head = 0;
   tail = 0;
}

static void print_par(void)
{
   printk(" Kernel ==> cal_ok = %d\n",cal_ok);
   printk(" Kernel ==> raw_max_x = %d\n",raw_max_x);
   printk(" Kernel ==> raw_max_y = %d\n",raw_max_y);
   printk(" Kernel ==> res_x = %d\n",res_x);
   printk(" Kernel ==> res_y = %d\n",res_y);
   printk(" Kernel ==> raw_min_x = %d\n",raw_min_x);
   printk(" Kernel ==> raw_min_y = %d\n",raw_min_y);
   printk(" Kernel ==> xyswap = %d\n",xyswap);
   printk(" Kernel ==> x_rev = %d\n",x_rev);
   printk(" Kernel ==> y_rev = %d\n",y_rev);
}

/* Allen */

static inline int pen_up(void)
{
   codec_write(CODEC_REG_ADC_CTL, 0);
   codec_write(CODEC_REG_TS_CTL, TSPX_POW | TSMX_POW | TSPY_GND | TSMY_GND);

   return codec_read(CODEC_REG_TS_CTL) & TSPX_LOW;
}

static void new_data(void)
{
   static TS_EVENT last_data = { 0, 0, 0, 0 };
   int diff0, diff1, diff2, diff3;

   if (cur_data.pressure)
   {
      if (sample < 3)
      {
	 samples[sample].x = cur_data.x;
	 samples[sample++].y = cur_data.y;
	 return;
      }
      sample = 0;

/* Check the variance between X samples (discard if not similar), then choose the closest pair */
      diff0 = abs(samples[0].x - samples[1].x);
      diff1 = abs(samples[1].x - samples[2].x);
      diff2 = abs(samples[2].x - cur_data.x);
      diff3 = abs(cur_data.x - samples[1].x);

      if (diff0 > XLIMIT || diff1 > XLIMIT || diff2 > XLIMIT || diff3 > XLIMIT)
	 return;

      if (diff1 < diff2)
      {
	 if (diff1 < diff3)
	    cur_data.x = (samples[1].x + samples[2].x) / 2;
	 else
	    cur_data.x = (cur_data.x + samples[1].x) / 2;
      }
      else
      {
	 if (diff2 < diff3)
	    cur_data.x = (samples[2].x + cur_data.x) / 2;
	 else
	    cur_data.x = (cur_data.x + samples[1].x) / 2;
      }

/* Do the same for Y */
      diff0 = abs(samples[0].y - samples[1].y);
      diff1 = abs(samples[1].y - samples[2].y);
      diff2 = abs(samples[2].y - cur_data.y);
      diff3 = abs(cur_data.y - samples[1].y);

      if (diff0 > YLIMIT || diff1 > YLIMIT || diff2 > YLIMIT || diff3 > YLIMIT)
	 return;

      if (diff1 < diff2)
      {
	 if (diff1 < diff3)
	    cur_data.y = (samples[1].y + samples[2].y) / 2;
	 else
	    cur_data.y = (cur_data.y + samples[1].y) / 2;
      }
      else
      {
	 if (diff2 < diff3)
	    cur_data.y = (samples[2].y + cur_data.y) / 2;
	 else
	    cur_data.y = (cur_data.y + samples[1].y) / 2;
      }
   }
   else
   {
/* Reset jitter detection on pen release */
      last_data.x = 0;
      last_data.y = 0;
   }

/* Jitter elimination */
/* Allen Mask */
//   if ((last_data.x || last_data.y) && abs(last_data.x - cur_data.x) <= 3 && abs(last_data.y - cur_data.y) <= 3)
//	return;
/* Allen */

   cur_data.millisecs = jiffies;
   last_data = cur_data;

   if (head != tail)
   {
      int last = head--;
      if (last < 0)
	 last = BUFSIZE - 1;
   }
   buf[head] = cur_data;

   if (++head == BUFSIZE)
      head = 0;
   if (head == tail && tail++ == BUFSIZE)
      tail = 0;
 
   if (fasync)
      kill_fasync(&fasync, SIGIO, POLL_IN);
   wake_up_interruptible(&queue);
}

static TS_EVENT get_data(void)
{
   int last = tail;

   if (++tail == BUFSIZE)
      tail = 0;
	
   return buf[last];
}

static void wait_for_action(void)
{
   state = PRESSED;
   sample = 0;
   codec_write(CODEC_REG_ADC_CTL, 0);
   codec_write(CODEC_REG_RISE_INT_ENABLE, 0);
   codec_write(CODEC_REG_FALL_INT_ENABLE, TSPX_INT);
   codec_write(CODEC_REG_TS_CTL, TSPX_POW | TSMX_POW | TSPY_GND | TSMY_GND | TSC_MODE_INT);
}

static void start_chain(void)
{
   state = P_DONE;
   codec_write(CODEC_REG_RISE_INT_ENABLE, ADC_INT);
   set_read_pressure();
}

static unsigned int sa1100_ts_poll(struct file *filp, struct poll_table_struct *wait)
{
   poll_wait(filp, &queue, wait);
   if (head != tail)
      return POLLIN | POLLRDNORM;
   return 0;
}

static ssize_t sa1100_ts_read(struct file *filp, char *buf, size_t count, loff_t *l)
{
   DECLARE_WAITQUEUE(wait, current);
   int i;
   TS_EVENT t;
   short out_buf[4];

   if (head == tail)
   {
      if (filp->f_flags & O_NONBLOCK)
	 return -EAGAIN;
      add_wait_queue(&queue, &wait);
      current->state = TASK_INTERRUPTIBLE;
      while ((head == tail) && !signal_pending(current))
      {
	 schedule();
	 current->state = TASK_INTERRUPTIBLE;
      }
      current->state = TASK_RUNNING;
      remove_wait_queue(&queue, &wait);
   }
   for (i = count; i >= sizeof(out_buf); i -= sizeof(out_buf), buf += sizeof(out_buf))
   {
      if (head == tail)
	 break;
      t = get_data();
      DPRINTK("p=%d, r_x= %d; r_y=%d\n",t.pressure, t.x, t.y);
      out_buf[0] = t.pressure;

/* Alen Add */
#if 0

#ifdef CONFIG_SA1100_ASSABET
      if (xyswap)
      {
	 out_buf[1] = (((raw_max_y - t.y)) * res_y) / (raw_max_y - raw_min_y);
	 out_buf[2] = (((t.x - raw_min_x)) * res_x) / (raw_max_x - raw_min_x);
      }
      else
      {
	 out_buf[1] = (((raw_max_x - t.x)) * res_x) / (raw_max_x - raw_min_x);
	 out_buf[2] = (((raw_max_y - t.y)) * res_y) / (raw_max_y - raw_min_y);
      }
#else
      if (xyswap)
      {
	 out_buf[1] = (((t.y - raw_min_y)) * res_y) / (raw_max_y - raw_min_y);
	 out_buf[2] = (((t.x - raw_min_x)) * res_x) / (raw_max_x - raw_min_x);
     if(out_buf[1] > res_y)
         out_buf[1] = res_y;
     if(out_buf[2] > res_x)
         out_buf[2] = res_x;
      }
      else
      {
	 out_buf[1] = (((t.x - raw_min_x)) * res_x) / (raw_max_x - raw_min_x);
	 out_buf[2] = (((t.y - raw_min_y)) * res_y) / (raw_max_y - raw_min_y);
      }
#endif
 

#else
      if (cal_ok)
      {
	 	out_buf[1] = (x_rev) ? ((raw_max_x - t.x) * res_x) / (raw_max_x - raw_min_x) :
		((t.x - raw_min_x) * res_x) / (raw_max_x - raw_min_x);
	 	out_buf[2] = (y_rev) ? ((raw_max_y - t.y) * res_y) / (raw_max_y - raw_min_y) :
		((t.y - raw_min_y) * res_y) / (raw_max_y - raw_min_y);

        /* For rotated screens */
        if(xyswap)
        {
            unsigned int temp;

            out_buf[2] = res_y - out_buf[2];

            temp = out_buf[1];
            out_buf[1] = out_buf[2];
            out_buf[2] = temp;
        }
        else
        {
            out_buf[1] = res_x - out_buf[1];
            out_buf[2] = res_y - out_buf[2];
        }

        if(out_buf[1] < 0)
            out_buf[1] = 0;

        if(out_buf[2] < 0)
            out_buf[2] = 0;

        if(xyswap)
        {
            if(out_buf[1] >= res_y)
                out_buf[1] = res_y - 1;

            if(out_buf[2] >= res_x)
                out_buf[2] = res_x - 1;
        }
        else
        {
            if(out_buf[2] >= res_y)
                out_buf[2] = res_y - 1;

            if(out_buf[1] >= res_x)
                out_buf[1] = res_x - 1;
        }
      }
      else
      {
	 	out_buf[1] = t.x;
	 	out_buf[2] = t.y;
      }
#endif
 
/* Allen */

      out_buf[3] = t.millisecs;
      DPRINTK("x= %d; y=%d\n",out_buf[1], out_buf[2]);
      copy_to_user(buf, &out_buf, sizeof(out_buf));
   }

   return count - i;
}

/* Forward declaration */
static void sa1100_ts_timer(unsigned long);

static int sa1100_ts_starttimer(void)
{
   in_timehandle++;
   init_timer(&timer);
   timer.function = sa1100_ts_timer;
   timer.expires = jiffies + HZ / 100;
   add_timer(&timer);
   return 0;
}

static void sa1100_ts_timer(unsigned long data)
{
   in_timehandle--;
   if (pen_up())
   {
      cur_data.pressure = 0;
      new_data();
      wait_for_action();
   }
   else
      start_chain();
}

static int sa1100_ts_fasync(int fd, struct file *filp, int on)
{
   int retval;

   retval = fasync_helper(fd, filp, on, &fasync);
   if (retval < 0)
      return retval;
   else
      return 0;
}

static int sa1100_ts_open(struct inode *inode, struct file *filp)
{
/* Allen Add */
   ts_clear();
/* Allen */

   MOD_INC_USE_COUNT;
   return 0;
}

static int sa1100_ts_release(struct inode *inode, struct file *filp)
{
/* Allen Add */
   ts_clear();
/* Allen */

   sa1100_ts_fasync(-1, filp, 0);
   MOD_DEC_USE_COUNT;
   return 0;
}

static void sa1100_ts_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
   ts_clear_codec_int();
   if (in_timehandle > 0)
      return;
   switch (state)
   {
      case PRESSED:
	 start_chain();
	 break;
      case P_DONE:
	 cur_data.pressure = read_adc();
	 codec_write(CODEC_REG_RISE_INT_ENABLE, ADC_INT);
	 codec_write(CODEC_REG_FALL_INT_ENABLE, 0);
	 set_read_x_pos();
	 state++;
	 break;
      case X_DONE:
	 cur_data.x = read_adc();
	 codec_write(CODEC_REG_RISE_INT_ENABLE, ADC_INT);
	 set_read_y_pos();
	 state++;
	 break;
      case Y_DONE:
	 cur_data.y = read_adc();
	 codec_write(CODEC_REG_RISE_INT_ENABLE, TSPX_INT);
	 codec_write(CODEC_REG_ADC_CTL, 0);
	 codec_write(CODEC_REG_TS_CTL, TSPX_POW | TSMX_POW |
	 TSPY_GND | TSMY_GND | TSC_MODE_INT);
	 state++;
	 new_data();
	 sa1100_ts_starttimer();
	 break;
      case RELEASED:
	 cur_data.pressure = 0;
	 new_data();
	 wait_for_action();
   }
}

static struct file_operations sa1100_ts_fops = {
   read: sa1100_ts_read,
   poll: sa1100_ts_poll,
   ioctl: sa1100_ts_ioctl,
   fasync: sa1100_ts_fasync,
   open: sa1100_ts_open,
   release: sa1100_ts_release,
};
/* Audio Codec Functions, etc */

/*
 * DMA callback functions
 */

static void audio_dmaout_done_callback(void *buf_id, int size)
{
	long int  temp=0;
	audio_buf_t *b = (audio_buf_t *) buf_id;
	/* 
	 * Current buffer is sent: wake up any process waiting for it.
	 */
	up(&b->sem);
        op_dma_chunk--;
	//printk("DMAOUT op_dma_chunk cnt=%ld\n", op_dma_chunk);
        if(op_dma_chunk<=0)
	{
		op_dma_chunk=0;
        	temp=codec_read(CODEC_REG_AUDIO_CTL_B);
        	if( temp & (1<<15) )
        	{
        		codec_write(CODEC_REG_AUDIO_CTL_B,(temp&(~(1<<15)) ));
			//printk("Val of CRACB %x \n", codec_read(CODEC_REG_AUDIO_CTL_B));
              		}
		}
        
}
static void audio_dmain_done_callback(void *buf_id, int size)
{
	audio_buf_t *b = (audio_buf_t *) buf_id;
	/* 
	 * Current buffer is full: set its size and wake up any 
	 * process waiting for it.
	 */
	b->size = size;
	up(&b->sem);
}

static int audio_clear_dma(audio_stream_t * s)
{
	sa1100_free_dma(s->dma_ch);
	return 0;
}

static int __init audio_init_dma(audio_stream_t * s, char *desc)
{
	int err;

	err = sa1100_request_dma(&s->dma_ch, desc);
	if (err)
		return err;
	if (s == &ucb1300_output_stream) {
		sa1100_dma_set_callback(s->dma_ch,
					audio_dmaout_done_callback);
		sa1100_dma_set_device(s->dma_ch, DMA_Ser4MCP0Wr);
	} else {
		sa1100_dma_set_callback(s->dma_ch,
					audio_dmain_done_callback);
		sa1100_dma_set_device(s->dma_ch, DMA_Ser4MCP0Rd);
	}
	return 0;
}




/* 
 * This function frees all buffers 
 */

static void audio_clear_buf(audio_stream_t * s)
{
        int temp;
	DPRINTK("audio_clear_buf\n");

	/* ensure DMA won't run anymore */
	sa1100_dma_flush_all(s->dma_ch);


   
	if (s->buffers) {
		int frag;
		for (frag = 0; frag < s->nbfrags; frag++) {
			if (!s->buffers[frag].master)
				continue;
			consistent_free(s->buffers[frag].start);
		}
		kfree(s->buffers);
		s->buffers = NULL;
	}

	s->buf_idx = 0;
	s->buf = NULL;
        op_dma_chunk=0;
	//printk("AUDIO_CLEAR_BUF op_dma_chunk cnt=%ld\n", op_dma_chunk);
        temp=codec_read(CODEC_REG_AUDIO_CTL_B);
        if(temp & (1<<15))
        {
        codec_write(CODEC_REG_AUDIO_CTL_B,(temp&(~(1<<15)) ));
	//printk("Val of CRACB %x \n", codec_read(CODEC_REG_AUDIO_CTL_B));
              }
        
}
/* This function allocates the buffer structure array and buffer data space
 * according to the current number of fragments and fragment size.
 */

static int audio_setup_buf(audio_stream_t * s)
{
	int frag;
	int dmasize = 0;
	char *dmabuf = 0;
	dma_addr_t dmaphys = 0;

	if (s->buffers)
		return -EBUSY;

	s->nbfrags = audio_nbfrags;
	s->fragsize = audio_fragsize;

	s->buffers = (audio_buf_t *)
	    kmalloc(sizeof(audio_buf_t) * s->nbfrags, GFP_KERNEL);
	if (!s->buffers)
		goto err;
	memset(s->buffers, 0, sizeof(audio_buf_t) * s->nbfrags);

	for (frag = 0; frag < s->nbfrags; frag++) {
		audio_buf_t *b = &s->buffers[frag];

		/*
		 * Let's allocate non-cached memory for DMA buffers.
		 * We try to allocate all memory at once.
		 * If this fails (a common reason is memory fragmentation), 
		 * then we allocate more smaller buffers.
		 */
		if (!dmasize) {
			dmasize = (s->nbfrags - frag) * s->fragsize;
			do {
				dmabuf = consistent_alloc(GFP_KERNEL,
							  dmasize,
							  &dmaphys);
				if (!dmabuf)
					dmasize -= s->fragsize;
			} while (!dmabuf && dmasize);
			if (!dmabuf)
				goto err;
			b->master = 1;
		}

		b->start = dmabuf;
		b->dma_addr = dmaphys;
		sema_init(&b->sem, 1);
		DPRINTK("buf %d: start %p dma %x\n", frag, b->start, b->dma_addr);

		dmabuf += s->fragsize;
		dmaphys += s->fragsize;
		dmasize -= s->fragsize;
	}

	s->buf_idx = 0;
	s->buf = &s->buffers[0];

	return 0;

      err:
	printk(AUDIO_NAME ": unable to allocate audio memory\n ");
	audio_clear_buf(s);
	return -ENOMEM;
}



static int audio_recording(audio_stream_t * s)
{
	int i;

	if (!s->buffers) {
		if (audio_setup_buf(s))
			return -ENOMEM;

		/*
		 * We must ensure there is an output stream at any time while 
		 * recording since this is how the UDA1341 gets its clock.
		 * So if there is no playback data to send, the output DMA will
		 * spin with all zeroes.
		 */
/*
		sa1100_dma_set_spin(ucb1300_output_stream.dma_ch,
				    (dma_addr_t) FLUSH_BASE_PHYS, 2048);
*/

		/* 
		 * Since we just allocated all buffers, we must send them to
		 * the DMA code before receiving data.
		 */
		for (i = 0; i < s->nbfrags; i++) {
			audio_buf_t *b = s->buf;
			down(&b->sem);
			sa1100_dma_queue_buffer(s->dma_ch, (void *) b,
						b->dma_addr, s->fragsize);
			NEXT_BUF(s, buf);
		}
	}
	return 0;
}
static int num_times=0;
static int audio_read(struct file *file, char *buffer,
		      size_t count, loff_t * ppos)
{
	char *buffer0 = buffer;
	audio_stream_t *s = &ucb1300_input_stream;
	int chunksize, ret = 0;

	DPRINTK("audio_read: count=%d\n", count);

	switch (file->f_flags & O_ACCMODE) {
	case O_RDONLY:
	case O_RDWR:
		break;
	default:
		return -EPERM;
	}


	if(num_times==0)
	{
             codec_write(CODEC_REG_AUDIO_CTL_B,codec_read(CODEC_REG_AUDIO_CTL_B)|(1<<14));
            //printk("CRACB %x \n",codec_read(CODEC_REG_AUDIO_CTL_B));
            num_times=1;
	}


	ret = audio_recording(s);
	if (ret)
		return ret;

	/* be sure to have a full sample byte count */
	count &= ~0x03;

	while (count > 0) {
		audio_buf_t *b = s->buf;

		/* Wait for a buffer to become full */
		if (file->f_flags & O_NONBLOCK) {
			ret = -EAGAIN;
			if (down_trylock(&b->sem))
				break;
		} else {
			ret = -ERESTARTSYS;
			if (down_interruptible(&b->sem))
				break;
		}

		/* Grab data from the current buffer */
		if (1) {
			chunksize = b->size;
			if (chunksize > count)
				chunksize = count;
			DPRINTK("read %d from %d\n", chunksize, s->buf_idx);
			if (copy_to_user(buffer,
					 b->start + s->fragsize - b->size,
					 chunksize)) {
				up(&b->sem);
				return -EFAULT;
			}
			b->size -= chunksize;
		} 

		buffer += chunksize;
		count -= chunksize;
		if (b->size > 0) {
			up(&b->sem);
			break;
		}

		/* Make current buffer available for DMA again */
		sa1100_dma_queue_buffer(s->dma_ch, (void *) b,
					b->dma_addr, s->fragsize);
		NEXT_BUF(s, buf);
	}

	if ((buffer - buffer0))
		ret = buffer - buffer0;
	DPRINTK("audio_read: return=%d\n", ret);
	return ret;
}
static int audio_write(struct file *file, const char *buffer,
		       size_t count, loff_t * ppos)
{
	const char *buffer0 = buffer;
	audio_stream_t *s = &ucb1300_output_stream;
	int chunksize, ret = 0;

	DPRINTK("audio_write: count=%d\n", count);

	if (!s->buffers && audio_setup_buf(s))
		return -ENOMEM;

	/* be sure to have a full sample byte count */
	count &= ~0x03;

	while (count > 0) {
		audio_buf_t *b = s->buf;

		/* Wait for a buffer to become free */
		if (file->f_flags & O_NONBLOCK) {
			ret = -EAGAIN;
			if (down_trylock(&b->sem))
				break;
		} else {
			ret = -ERESTARTSYS;
			if (down_interruptible(&b->sem))
				break;
		}

		/* Feed the current buffer */
		chunksize = s->fragsize - b->size;
		if (chunksize > count)
			chunksize = count;
		DPRINTK("write %d to %d\n", chunksize, s->buf_idx);
		if (copy_from_user(b->start + b->size, buffer, chunksize)) {
			up(&b->sem);
			return -EFAULT;
		}
		buffer += chunksize;
		count -= chunksize;
		b->size += chunksize;
		if (b->size < s->fragsize) {
			up(&b->sem);
			break;
		}

		/* Send current buffer to dma */
		sa1100_dma_queue_buffer(s->dma_ch, (void *) b,
					b->dma_addr, b->size);
		b->size = 0;	/* indicate that the buffer has been sent */
		NEXT_BUF(s, buf);
		op_dma_chunk++;
		//printk("op_dma_chunk cnt=%ld\n", op_dma_chunk);
                if(op_dma_chunk==1)
                {
   			codec_write(CODEC_REG_AUDIO_CTL_B,(1<<15)|0x0);
			//printk("Val of CRACB %x \n", codec_read(CODEC_REG_AUDIO_CTL_B));
                  }
	}

	if ((buffer - buffer0))
		ret = buffer - buffer0;
	DPRINTK("audio_write: return=%d\n", ret);
	return ret;
}




static int audio_ioctl(struct inode *inode, struct file *file,
                       uint cmd, ulong arg)
{
	long val,temp,scratch;
	static long int dsp_sfmt=0;
	
	switch(cmd){
		case SNDCTL_DSP_SPEED:
				get_user(val, (long *) arg);
				if( (val < 8000) || (val > 62500) )
				   return -EINVAL;
			 	temp=12000000/(32*val);	
				//printk("Val of SNDCTL_DSP_SPEED %ld\n",temp);
 
/* Disable the audio or telecom enable stuff here before changing the
rates */
   				scratch=codec_read(CODEC_REG_AUDIO_CTL_B);
				if( (scratch&(1<<15)) || (scratch&(1<<14)) )
				  {
					printk("Record or Playback in \
progress\n");
					return -EINVAL;
					}
				else{
					//printk("Enabling diff asd..\n");
					Ser4MCCR0 = ( Ser4MCCR0 & 0xffffff00 ) | temp;
					codec_write(CODEC_REG_AUDIO_CTL_B,temp);
            				//printk("CRACB %x \n",codec_read(CODEC_REG_AUDIO_CTL_B));
					}

				break;
		case SNDCTL_DSP_SETFRAGMENT:
				break;
		case SNDCTL_DSP_SETFMT:
				get_user(dsp_sfmt, (long *) arg);
				break;
		case SNDCTL_DSP_GETFMTS:
				put_user(dsp_sfmt, (long *) arg);
				break;
		case SNDCTL_DSP_STEREO:
                		get_user(val, (long *) arg);
				//printk("Val of DSP Stereo %lu\n",val);
				if(val==2 || val > 2)
					return -EINVAL;
				break;
		case SNDCTL_DSP_RESET:
				break;
		default:	
				   return -EINVAL;
	}

return 0;
}



static int audio_open(struct inode *inode, struct file *file)
{
	DPRINTK("audio_open\n");

	if (audio_refcount == 0) {
		audio_rate = AUDIO_RATE_DEFAULT;
		audio_fragsize = AUDIO_FRAGSIZE_DEFAULT;
		audio_nbfrags = AUDIO_NBFRAGS_DEFAULT;
		audio_clear_buf(&ucb1300_output_stream);
	}
	audio_refcount++;

	MOD_INC_USE_COUNT;
	return 0;
}

static int audio_release(struct inode *inode, struct file *file)
{
	//printk("audio_release start\n");

	audio_refcount--;
	if (audio_refcount == 0) {
		//printk("AUDIO_RELEASE pos1 \n");
		//audio_sync(file);
		//printk("AUDIO_RELEASE pos2 \n");
		sa1100_dma_set_spin(ucb1300_output_stream.dma_ch, 0, 0);
		//printk("AUDIO_RELEASE pos3 \n");
		audio_clear_buf(&ucb1300_output_stream);
	}

	//printk("audio_release end\n");
	MOD_DEC_USE_COUNT;
	return 0;
}

static struct file_operations UCB1300_dsp_fops = {
	llseek:		NULL,
	write:		audio_write,
	read:		audio_read,
	poll:		NULL,
	ioctl:		audio_ioctl,
	open:		audio_open,
	release:	audio_release
};

/* Audio Codec Functions, etc */
int __init sa1100_ts_init(void)
{
   unsigned int irq;

/*	
   raw_max_x = 944;
   raw_max_y = 944;
   raw_min_x = 70;
   raw_min_y = 70;
*/	
/*
   raw_max_x = 896;
   raw_max_y = 886;
   raw_min_x = 187;
   raw_min_y = 127;
   xyswap = 0;
*/
	
   res_x = 320;
   res_y = 240;

   head = 0;
   tail = 0;

/* Allen Add */
   cal_ok = 1;
   x_rev = 0;
   y_rev = 0;
/* Allen */

   init_waitqueue_head(&queue);

/*
	Not needed now cos this is done in audio_assabet_ucb1300_init
      BCR_set(BCR_CODEC_RST);
      Ser4MCCR0 = 0x10000;
*/
      audio_assabet_ucb1300_init();
      GPDR &= ~GPIO_GPIO5;
      set_GPIO_IRQ_edge(GPIO_GPIO5, GPIO_BOTH_EDGES);

/* Initialize the touchscreen controller */
   codec_write(CODEC_REG_ADC_CTL, 0);
   codec_write(CODEC_REG_RISE_INT_ENABLE, 0);
   codec_write(CODEC_REG_FALL_INT_ENABLE, 0);
   codec_write(CODEC_REG_IO_DIRECTION, 0x200);
   ts_clear_codec_int();
   register_chrdev(TS_MAJOR, TS_NAME, &sa1100_ts_fops);

	irq = 5;
	

   if (request_irq(irq, sa1100_ts_interrupt, SA_INTERRUPT | SA_SHIRQ, TS_NAME, dev_id))
      printk("sa1100_ts_init: failed to register IRQ\n");

   wait_for_action();
   printk("sa1100 touch screen driver initialized\n");

/* Initialize the audio part now */
	if (audio_init_dma(&ucb1300_output_stream, "UCB1300 out") ||
 audio_init_dma(&ucb1300_input_stream, "UCB1300 in")  ) {
		audio_clear_dma(&ucb1300_output_stream);
		audio_clear_dma(&ucb1300_input_stream);
		printk("Audio_init FAILED\n");
	}
	audio_dev_dsp = register_sound_dsp(&UCB1300_dsp_fops, -1);
	printk(AUDIO_NAME_VERBOSE " initialized\n");

   return 0;
}

void __exit sa1100_ts_cleanup(void)
{
   unsigned int irq;
	irq=5;
	

   codec_write(CODEC_REG_ADC_CTL, 0);
   codec_write(CODEC_REG_RISE_INT_ENABLE, 0);
   codec_write(CODEC_REG_FALL_INT_ENABLE, 0);
   free_irq(irq, dev_id);
   if (in_timehandle)
      del_timer(&timer);
   unregister_chrdev(TS_MAJOR, TS_NAME);

   printk("sa1100 touch screen driver removed\n");
}

module_init(sa1100_ts_init);
module_exit(sa1100_ts_cleanup);

MODULE_PARM(raw_max_x,"i");
MODULE_PARM(raw_max_y,"i");
MODULE_PARM(raw_min_x,"i");
MODULE_PARM(raw_min_y,"i");
MODULE_PARM(xyswap,"i");

static inline void audio_assabet_ucb1300_init(void)
{
   BCR_set(BCR_CODEC_RST);
   BCR_set(BCR_AUDIO_ON);
   BCR_clear(BCR_STEREO_LB| BCR_QMUTE | BCR_SPK_OFF);

   Ser4MCCR0=0x0;
/* 9 for 44.1
   0x17 for 16K
   0x2e for 8K
*/
   Ser4MCCR0 = 0x200009;
   Ser4MCCR1= 0x000000;
   printk("Val of Ser4mccr0 %x \n", Ser4MCCR0);
   printk("Val of MCSR %x \n", Ser4MCSR);
   printk("Val of DDAR_ser4mcp0wr %x \n", DDAR_Ser4MCP0Wr);

/* Enable the mce */
   Ser4MCCR0 |= MCCR0_MCE;
   printk("Val of Ser4mccr0 %x \n", Ser4MCCR0);

   printk("Val of CRIOD %x \n", codec_read(CODEC_REG_IO_DIRECTION));
   codec_write(CODEC_REG_IO_DIRECTION,(1<<15));
   printk("Val of CRIOD %x \n", codec_read(CODEC_REG_IO_DIRECTION));

   printk("Val of CRMode %x \n", codec_read(CODEC_REG_MODE));
   codec_write(CODEC_REG_MODE,(1<<12));
   printk("Val of CRIOD after dyna set %x \n", codec_read(CODEC_REG_MODE));
   codec_write(CODEC_REG_AUDIO_CTL_A,0);
   codec_write(CODEC_REG_AUDIO_CTL_A,0x09);
   printk("Val of CRACA %x \n", codec_read(CODEC_REG_AUDIO_CTL_A));

   codec_write(CODEC_REG_AUDIO_CTL_B,0);

/* Enable the codec device only after first write */
/*
   codec_write(CODEC_REG_AUDIO_CTL_B,(1<<15)|0x0);
   printk("Val of CRACB %x \n", codec_read(CODEC_REG_AUDIO_CTL_B));
*/
}
