#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/interrupt.h>
#include <linux/errno.h>
#include <linux/kbd_ll.h>

#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/hardware.h>
#include <asm/semaphore.h>
#include <asm/dma.h>
#include <asm/irq.h>

//#define DEBUG

#define KEYPAD_NAME		        "skpd"
#define KEYPAD_NAME_VERBOSE	    "Simputer Keypad driver"
#define KEYPAD_VERSION_STRING	"version 0.1"

#define KEYPAD_DEVID            "simp_keypad"


static void keypad_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    int gpio,level;
    unsigned char scancode;

#ifdef DEBUG
	printk(KERN_CRIT"Calling keypad interrupt, irq=%d\n",irq);
#endif

    gpio = irq - 32 + 11;

    switch(gpio)
    {
        case 11:
            scancode = 30;
            break;

        case 12:
            scancode = 48;
            break;

        case 13:
            scancode = 46;
            break;
            
        case 14:
            scancode = 32;
            break;

        case 15:
            scancode = 18;
            break;

        default:
            printk(KERN_ERR "Unknown key pressed ??\n");
            return;
    }

    level = (GPLR & (1 << gpio));

    if( level)
    {
#ifdef DEBUG
        printk(KERN_CRIT"high\n");
#endif
        /* Key pressed */
        handle_scancode(scancode, 0);
    }
    else
    {
#ifdef DEBUG
        printk(KERN_CRIT"low\n");
#endif
        /* Key released */
        handle_scancode(scancode, 1);
    }

    return ;
}

int simp_keypad_translate(unsigned char scancode, unsigned char *keycode_p, char raw_mode)
{
#ifdef DEBUG
    printk(KERN_CRIT "simp_keypad_translate : scancode=%d\n", scancode);
#endif
    *keycode_p = scancode & 0x7f;
    return 1;
}

static struct file_operations keypad_fops = {
   read: NULL,
   poll: NULL,
   ioctl: NULL,
   fasync: NULL,
   open: NULL,
   release: NULL,
};

static int major_nr;

int __init keypad_init(void)
{
    int irq; 
     
    GPDR &= ~(GPIO_GPIO11|GPIO_GPIO12|GPIO_GPIO13|GPIO_GPIO14|GPIO_GPIO15);

    /*
    set_GPIO_IRQ_edge(GPIO_GPIO11, GPIO_FALLING_EDGE);
    set_GPIO_IRQ_edge(GPIO_GPIO12, GPIO_FALLING_EDGE);
    set_GPIO_IRQ_edge(GPIO_GPIO13, GPIO_FALLING_EDGE);
    set_GPIO_IRQ_edge(GPIO_GPIO14, GPIO_FALLING_EDGE);
    set_GPIO_IRQ_edge(GPIO_GPIO15, GPIO_FALLING_EDGE);
    */

    set_GPIO_IRQ_edge(GPIO_GPIO11, GPIO_BOTH_EDGES);
    set_GPIO_IRQ_edge(GPIO_GPIO12, GPIO_BOTH_EDGES);
    set_GPIO_IRQ_edge(GPIO_GPIO13, GPIO_BOTH_EDGES);
    set_GPIO_IRQ_edge(GPIO_GPIO14, GPIO_BOTH_EDGES);
    set_GPIO_IRQ_edge(GPIO_GPIO15, GPIO_BOTH_EDGES);

   	major_nr = register_chrdev(0, KEYPAD_DEVID, &keypad_fops);

    if(major_nr <= 0)
    {
        printk(KERN_ERR"keypad_init: failed to register device\n");
        return 1;
    }

    for(irq=IRQ_GPIO11; irq <IRQ_GPIO16; irq ++)
        if (request_irq(irq, keypad_interrupt, SA_INTERRUPT | SA_SHIRQ, KEYPAD_DEVID , KEYPAD_DEVID))
        {
            printk(KERN_ERR"keypad_init: failed to register handler for irq: %d\n",irq);
            unregister_chrdev(major_nr, KEYPAD_DEVID);
            return 1;
        }


#ifdef DEBUG
	printk(KERN_INFO KEYPAD_NAME_VERBOSE " loaded\n");
#endif

	return 0;
}

module_init(keypad_init);

void __exit keypad_exit(void)
{
    int irq;


    for(irq=IRQ_GPIO11; irq <IRQ_GPIO16; irq ++)
        free_irq(irq, KEYPAD_DEVID);

    unregister_chrdev(major_nr, KEYPAD_DEVID);
#ifdef DEBUG
	printk(KERN_INFO KEYPAD_NAME_VERBOSE " unloaded\n");
#endif
}

module_exit(keypad_exit);

void simp_keypad_leds(unsigned char leds)
{
}

void simp_keypad_init_hw(void)
{
}

void simp_keypad_enable_irq(void)
{
}

void simp_keypad_disable_irq(void)
{
}

