#include <linux/types.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <asm/hardware.h>
#include <asm/irq.h>
#include <linux/string.h>
#include "ucb1200_ts.h"

static char *dev_id = "ucb1200";
static int state, Reg10;

#define DPRINTK(x...) printk(KERN_NOTICE x)


/* Useful operations */
static inline ushort codec_read(ushort addr)
{
    ulong flags;
    ushort data;

//    DPRINTK("inside codec_read: %x\n", addr);
    save_flags_cli(flags);
    while (!(Ser4MCSR & MCSR_CWC));
    Ser4MCDR2 = ((addr & 0xf) << FShft(MCDR2_ADD)) | MCDR2_Rd;

    while (!(Ser4MCSR & MCSR_CRC));
    data = Ser4MCDR2 & 0xffff;
    restore_flags(flags);
    return data;
}

static inline void codec_write(ushort addr, ushort data)
{
    ulong flags;
//    DPRINTK("inside codec_write: %x %x\n", addr, data);
    save_flags_cli(flags);
    while (!(Ser4MCSR & MCSR_CWC));
    Ser4MCDR2 = ((addr & 0xf) << FShft(MCDR2_ADD)) | MCDR2_Wr | (data & 0xffff);
    restore_flags(flags);
}

static inline void ts_clear_codec_int(void)
{
   int s;
   DPRINTK("\tinside ts_clear_codec_int\n");
/* Clear interrupts at codec  */
   if ((s = codec_read(CODEC_REG_INT_STATUS) & (TSPX_INT|TSMX_INT|ADC_INT)))
   {
      codec_write(CODEC_REG_INT_STATUS, 0);
      codec_write(CODEC_REG_INT_STATUS, s);
   }
}


static struct file_operations sa1100_ts_fops = {};

static void sa1100_ts_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	
   DPRINTK("\tinside ts_interrupt\n");
      ts_clear_codec_int();
   codec_write(9, TSPX_POW | TSMX_POW | TSPY_GND | TSMY_GND | TSC_MODE_PRESSURE | TSC_BIAS_ENA);
// codec_write(9, TSPX_POW | TSMX_POW | TSPY_GND | TSMY_GND | TSC_MODE_PRESSURE | TSC_BIAS_ENA);
   codec_write(CODEC_REG_TS_CTL, TSPY_POW | TSMY_GND | TSC_MODE_POSITION | TSC_BIAS_ENA);
// codec_write(CODEC_REG_RISE_INT_ENABLE, ADC_INT);
   codec_write(CODEC_REG_FALL_INT_ENABLE, 0);			
   codec_write(CODEC_REG_ADC_CTL, Reg10 = ADC_INPUT_TSPX | ADC_ENA| ADC_SYNC_ENA);
   udelay(50);	
   codec_write(CODEC_REG_ADC_CTL, Reg10 | ADC_START);
   codec_write(CODEC_REG_IO_DATA, codec_read(CODEC_REG_IO_DATA) | 0x200);
   codec_write(CODEC_REG_ADC_CTL, Reg10);
   codec_write(CODEC_REG_IO_DATA, codec_read(CODEC_REG_IO_DATA) & ~0x200);
   udelay(50);		
   DPRINTK("ADC %x\n",codec_read(CODEC_REG_ADC_DATA));
   codec_write(CODEC_REG_RISE_INT_ENABLE, 0x0);
   codec_write(CODEC_REG_FALL_INT_ENABLE, 0x3000); 	
   codec_write(CODEC_REG_ADC_CTL, 0);
   codec_write(CODEC_REG_TS_CTL, TSPX_POW | TSMX_POW | TSPY_GND | TSMY_GND | TSC_MODE_INT);
   DPRINTK("\toutside ts_interrupt\n");

}

int __init sa1100_ts_init(void)
{
      unsigned int irq;
      BCR_set(BCR_CODEC_RST); 
      Ser4MCCR0 = 0x10000;
      GPDR &= ~GPIO_GPIO5;	   
      set_GPIO_IRQ_edge(GPIO_GPIO5,3);	   

/* Initialize the touchscreen controller */
   codec_write(CODEC_REG_ADC_CTL, 0);
   codec_write(CODEC_REG_RISE_INT_ENABLE, 0);
   codec_write(CODEC_REG_FALL_INT_ENABLE, 0);
   codec_write(CODEC_REG_IO_DIRECTION, 0x200);
   ts_clear_codec_int();
	
   register_chrdev(TS_MAJOR, TS_NAME, &sa1100_ts_fops);
   irq = IRQ_GPIO5;	
   state=0; 
   if (request_irq(irq, sa1100_ts_interrupt, SA_INTERRUPT | SA_SHIRQ, TS_NAME, dev_id))
      printk("sa1100_ts_init: failed to register IRQ\n");
   codec_write(CODEC_REG_RISE_INT_ENABLE, 0x3000);
   codec_write(CODEC_REG_FALL_INT_ENABLE, 0x3000);
   codec_write(CODEC_REG_ADC_CTL, 0);
   codec_write(CODEC_REG_TS_CTL, TSPX_POW | TSMX_POW | TSPY_GND | TSMY_GND | TSC_MODE_INT);
	
   printk("sa1100 touch screen driver initialized\n");
   return 0;
}

void __exit sa1100_ts_cleanup(void)
{
   unsigned int irq;

   irq = IRQ_GPIO5;
   codec_write(CODEC_REG_ADC_CTL, 0);
   codec_write(CODEC_REG_RISE_INT_ENABLE, 0);
   codec_write(CODEC_REG_FALL_INT_ENABLE, 0);
   free_irq(irq, dev_id);
   unregister_chrdev(TS_MAJOR, TS_NAME);
   printk("sa1100 touch screen driver removed\n");
}

module_init(sa1100_ts_init);
module_exit(sa1100_ts_cleanup);
